VERSION = "v0.7.1"
ANALYSIS_PIPELINE_VERSION = "v0.9.3"

MISSENSE_VARIANT = "missense_variant"
SYNONYMOUS_VARIANT = "synonymous_variant"
INFRAME_DELETION = "inframe_deletion"
INFRAME_INSERTION = "inframe_insertion"
DISRUPTIVE_INFRAME_DELETION = "disruptive_inframe_deletion"
CONSERVATIVE_INFRAME_DELETION = "conservative_inframe_deletion"
DISRUPTIVE_INFRAME_INSERTION = "disruptive_inframe_insertion"
CONSERVATIVE_INFRAME_INSERTION = "conservative_inframe_insertion"
